{% extends "!layout.html" %}
{% set script_files = script_files + ["_static/togglediv.js", "_static/mathconf.js"] %}
{% block rootrellink %}
    <li><a href="http://kwant-project.org/">
        <img src="{{ pathto('_static/kwant_logo.png', 1) }}"
        style="vertical-align: middle; margin-top: -4px" alt="Kwant logo"/></a>
      {{ reldelim1 }}</li>
    <li><a href="{{ pathto('index') }}">{{ shorttitle }}</a>{{ reldelim1 }}</li>
{% endblock %}

  </head>
  <body>
{%- if website_deploy %}
    <!-- Piwik -->
    <script type="text/javascript">
      var _paq = _paq || [];
      _paq.push(["setDocumentTitle", document.domain + "/" + document.title]);
      _paq.push(["setCookieDomain", "*.kwant-project.org"]);
      _paq.push(["trackPageView"]);
      _paq.push(["enableLinkTracking"]);

      (function() {
      var u=(("https:" == document.location.protocol) ? "https" : "http") + "://piwik.kwant-project.org/";
      _paq.push(["setTrackerUrl", u+"piwik.php"]);
      _paq.push(["setSiteId", "1"]);
      var d=document, g=d.createElement("script"), s=d.getElementsByTagName("script")[0]; g.type="text/javascript";
      g.defer=true; g.async=true; g.src=u+"piwik.js"; s.parentNode.insertBefore(g,s);
      })();
    </script>
    <noscript><img src="http://piwik.kwant-project.org/piwik.php?idsite=1&amp;rec=1" style="border:0" alt="" /></noscript>
    <!-- End Piwik Code: Analytics are used by us to improve the site, and not sent to anyone. -->
{%- endif %}
  </body>
</html>
